package graph;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class Graph {
	
	private List<Vertex> vertices = new ArrayList<Vertex>();
	private List<Edge> edges = new ArrayList<Edge>();
	private Set<Vertex> settledNodes = new HashSet<Vertex>();
	private Set<Vertex> unsettledNodes = new HashSet<Vertex>();
	private Map<Vertex, Vertex> predecessors = new HashMap<Vertex, Vertex>();
	private Map<Vertex, Integer> distance = new HashMap<Vertex, Integer>();
	private int[][] matrix;

	public Graph() {
		
	}

	public List<Vertex> getVertices() {
		return vertices;
	}

	public List<Edge> getEdges() {
		return edges;
	}

	public int[][] getMatrix() {
		return matrix;
	}

	public void setMatrix(int[][] matrix) {
		this.matrix = matrix;
	}

	public void addEdge(int firstVertex, int secondVertex, int id, int weight) {
		
		Edge current = new Edge(vertices.get(firstVertex), vertices.get(secondVertex), id, weight);
		edges.add(current);
	}

	public void addVertices(int number) {

		for(int i = 0; i < number; i++) {
			
			Vertex current = new Vertex(i);
			vertices.add(current);
		}
	}

	public void execute(Vertex source) {
		
		distance.put(source, 0);
		unsettledNodes.add(source);
		
		while(unsettledNodes.size() > 0) {
			
			Vertex node = getMinimum(unsettledNodes);
			settledNodes.add(node);
			unsettledNodes.remove(node);
			findMinimalDistances(node);
		}
	}

	private void findMinimalDistances(Vertex node) {
		
		List<Vertex> adjacentNodes = getNeighbours(node);
		
		for(Vertex target : adjacentNodes) {
			if(getShortestDistance(target) > getShortestDistance(node) + getDistance(node, target)) {
				
				distance.put(target, getShortestDistance(node) + getDistance(node, target));
				predecessors.put(target, node);
				unsettledNodes.add(target);
			}
		}
	}

	private int getDistance(Vertex node, Vertex target) {
		
		for(Edge edge : edges) {
			if(edge.getFirstVertex().equals(node) && edge.getSecondVertex().equals(target)) {
				return edge.getWeight();
			}
		}
		
		return -1;
	}

	private List<Vertex> getNeighbours(Vertex node) {
		
		List<Vertex> neighbours = new ArrayList<Vertex>();
		
		for(Edge edge : edges) {
			if(edge.getFirstVertex().equals(node) && !isSettled(edge.getSecondVertex())) {
				
				neighbours.add(edge.getSecondVertex());
			}
		}
		
		return neighbours;
	}

	private Vertex getMinimum(Set<Vertex> vertices) {
		
		Vertex minimum = null;
		
		for(Vertex vertex : vertices) {
			if(minimum == null) {
				
				minimum = vertex;
			} else {
				if(getShortestDistance(vertex) < getShortestDistance(minimum)) {
					
					minimum = vertex;
				}
			}
		}
		
		return minimum;
	}

	private boolean isSettled(Vertex vertex) {
		
		return settledNodes.contains(vertex);
	}

	private int getShortestDistance(Vertex destination) {
		
		Integer current = distance.get(destination);
		
		if(current == null) {
			
			return Integer.MAX_VALUE;
		} else {
			
			return current;
		}
	}

	public LinkedList<Vertex> getPath(Vertex target) {
		
		LinkedList<Vertex> path = new LinkedList<Vertex>();
		Vertex step = target;
		
		if(predecessors.get(step) == null) {
			
			return null;
		}
		
		path.add(step);
		
		while(predecessors.get(step) != null) {
			
			step = predecessors.get(step);
			path.add(step);
		}
		
		Collections.reverse(path);
		
		return path;
	}
}